# Copyright (c) 2025, Huawei Technologies Co., Ltd
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0  (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from copy import deepcopy

import pandas as pd

from msprof_analyze.cluster_analyse.common_func.table_constant import TableConstant
from msprof_analyze.cluster_analyse.common_func.utils import UnionFind
from msprof_analyze.prof_common.logger import get_logger
from msprof_analyze.prof_common.constant import Constant
from msprof_analyze.prof_common.database_service import DatabaseService
from msprof_analyze.prof_common.file_manager import FileManager

logger = get_logger()


class StageInfoAnalysis:

    def __init__(self, param: dict):
        self.cluster_analysis_output_path = param.get(Constant.CLUSTER_ANALYSIS_OUTPUT_PATH, "")
        self.cluster_analysis_output_dir = os.path.join(self.cluster_analysis_output_path,
                                                        Constant.CLUSTER_ANALYSIS_OUTPUT)
        self.data_type = param.get(Constant.DATA_TYPE)
        self.simplified_mode = param.get(Constant.DATA_SIMPLIFICATION)
        self.communication_data_dict = param.get(Constant.COMM_DATA_DICT, {})
        self.collective_group_dict = {}
        self.p2p_link = []
        self.p2p_union_group = []
        self.stage_group = []

    def run(self):
        if not self.prepare_data():
            return []
        self.generate_p2p_union_group()
        self.generate_stage_group()
        return self.stage_group

    def prepare_data(self):
        if Constant.KEY_COMM_GROUP_PARALLEL_INFO in self.communication_data_dict:
            comm_group_df = pd.DataFrame(self.communication_data_dict.get(Constant.KEY_COMM_GROUP_PARALLEL_INFO))
        else:
            comm_group_df = self.load_communication_group_df()
        return self.extract_infos(comm_group_df)

    def load_communication_group_df(self):
        if not os.path.exists(self.cluster_analysis_output_path):
            logger.warning(f"StageInfoAnalysis: {self.cluster_analysis_output_path} not exist!")
            return None
        return self.load_communication_group_df_for_text() if self.data_type == Constant.TEXT else (
            self.load_communication_group_df_for_db())

    def load_communication_group_df_for_text(self):
        # check file exist
        communication_group_json = os.path.join(self.cluster_analysis_output_dir, Constant.COMMUNICATION_GROUP_JSON)
        if not os.path.exists(communication_group_json):
            logger.warning(f"{communication_group_json} not exists!")
            return None
        # read comm_group_parallel_info from communication_group.json
        group_data = FileManager.read_json_file(communication_group_json)
        if Constant.KEY_COMM_GROUP_PARALLEL_INFO not in group_data:
            logger.warning(f"{Constant.KEY_COMM_GROUP_PARALLEL_INFO} not in {Constant.COMMUNICATION_GROUP_JSON}")
            return None
        # convert to dataframe
        comm_group_df = pd.DataFrame(group_data.get(Constant.KEY_COMM_GROUP_PARALLEL_INFO))
        comm_group_df[TableConstant.RANK_SET] = comm_group_df[TableConstant.RANK_SET].apply(set)
        return comm_group_df

    def load_communication_group_df_for_db(self):
        # load data from cluster_analysis.db
        if not os.path.exists(self.cluster_analysis_output_dir):
            logger.warning("rank %s db path %s does not exist.", self.cluster_analysis_output_path)
        cluster_analysis_db = os.path.join(self.cluster_analysis_output_dir,
                                           Constant.DB_CLUSTER_COMMUNICATION_ANALYZER)
        data_service = DatabaseService(cluster_analysis_db, {})
        if self.simplified_mode:
            table_communication_group = Constant.TABLE_COMMUNICATION_GROUP_MAPPING
        else:
            table_communication_group = Constant.TABLE_COMMUNICATION_GROUP
        data_service.add_table_for_query(table_communication_group)
        data_dict = data_service.query_data()
        comm_group_df = data_dict.get(table_communication_group, None)
        if comm_group_df is None or comm_group_df.empty:
            logger.error(f"There is no {table_communication_group} data in {cluster_analysis_db}.")
            return None

        # process rank_set
        comm_group_df[TableConstant.RANK_SET] = comm_group_df[TableConstant.RANK_SET].apply(
            lambda s: set(map(int, s.strip('()').split(','))))
        return comm_group_df

    def extract_infos(self, comm_group_df):
        if comm_group_df is None:
            return False
        self.collective_group_dict = \
            comm_group_df[comm_group_df[TableConstant.TYPE] == Constant.COLLECTIVE].set_index(TableConstant.GROUP_NAME)[
                TableConstant.RANK_SET].to_dict()
        pp_df = comm_group_df[comm_group_df[TableConstant.TYPE] == Constant.P2P]
        pp_df = pp_df[pp_df[TableConstant.PG_NAME].str.lower().str.startswith('pp', na=False)]
        self.p2p_link = pp_df[TableConstant.RANK_SET].to_list()
        return len(self.p2p_link) > 0

    def generate_p2p_union_group(self):
        self.p2p_link.sort(key=lambda x: min(x))
        while self.p2p_link:
            union_set = deepcopy(self.p2p_link[0])
            rm_list = [self.p2p_link[0]]
            for link_rank_set_x in self.p2p_link[1:]:
                if UnionFind.is_connected(link_rank_set_x, union_set):
                    union_set = union_set.union(link_rank_set_x)
                    rm_list.append(link_rank_set_x)
            self.p2p_union_group.append(union_set)
            self.p2p_link = [element for element in self.p2p_link if element not in rm_list]

    def generate_stage_group(self):
        stage_group = {}
        for _, rank_set in self.collective_group_dict.items():
            if not self.whether_valid_comm_group(rank_set):
                continue
            unioned_set = set()
            remove_key = []
            for first_rank, stage in stage_group.items():
                if UnionFind.is_connected(rank_set, stage):
                    unioned_set = UnionFind.union(rank_set, stage, unioned_set)
                    remove_key.append(first_rank)
            if unioned_set:
                for key in remove_key:
                    del stage_group[key]
                stage_group[min(unioned_set)] = unioned_set
            else:
                stage_group[min(rank_set)] = rank_set
        first_rank_sort_list = sorted(first_rank for first_rank in stage_group)
        self.stage_group = [list(stage_group.get(first_rank, {})) for first_rank in first_rank_sort_list]

    def whether_valid_comm_group(self, rank_set: set):
        """
        while distinguish which communication group should be used to infer stage info, these group should be ignored:
            1. group can not include more than 1 rank in every single p2p group
        """
        for p2p_rank_set in self.p2p_union_group:
            if len(rank_set.intersection(p2p_rank_set)) > 1:
                return False
        return True
